# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
# Copyright (c) 2010 Citrix Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A not-so-virtual-machine manager
"""

import json
import os
import paramiko

from nova import context
from nova import db
from nova import exception
from nova import flags
from nova import log as logging
from nova import utils
from nova.auth import manager
from nova.compute import instance_types
from nova.compute import power_state
from nova.virt import disk
from nova.virt import images

LOG = logging.getLogger('nova.virt.notsovm')

FLAGS = flags.FLAGS
#TODO(justinsb): These should really be moved to the database...
flags.DEFINE_string('notso.hosts', '00:1c:c0:da:00:9e=10.2.1.230',
                    'Comma-separated list of MAC=IP values')
flags.DEFINE_string('notso_privatekey', '',
                    'Filename of private key to use for SSH authentication')
flags.DEFINE_string('pxeboot_config_template',
                    utils.abspath('virt/pxeboot.config.template'),
                    'Template file for pxe boot')
flags.DEFINE_string('pxeboot_nfsserver',
                    '',
                    'NFS server path to use with PXE boot')


def get_connection(read_only):
    return NotsovmHost(read_only)


class PhysicalMachineNetworkInfo(object):
    def __init__(self, string_repr):
        print string_repr
        data = json.loads(string_repr)
        self.data = data
        self.primary_ip = self.data[0]['ip']
        self.primary_mac = self.data[0]['mac']

class ComputeHostBase(object):
    def update_available_resource(self, ctxt, host):
        """This method is supported only by libvirt."""
        return


class PxeBootManager(object):
    def __init__(self):
        pass

    def configure_pxe(self, machine, net_info, boot_target):
        with open(FLAGS.pxeboot_config_template) as f:
            conf = f.read() % { 'boot_target': boot_target,
                                'nfs_root': FLAGS.pxeboot_nfsserver }
        pxe_file = '01-' + net_info.primary_mac.replace(':', '-')
        pxe_path = ('/var/lib/tftpboot/pxelinux.cfg/%s' %
                    (pxe_file))
        with open(pxe_path, 'w') as f:
            f.write(conf)


class NotsovmHost(ComputeHostBase):
    def __init__(self, read_only):
        self.pxe_manager = PxeBootManager()

    def init_host(self, host):
        # TODO(justinsb): Adopt existing VM's running here
        LOG.warning(_('notsovm init_host stub-implemented'))
        pass

    def _db_query_all(self, context):
        pms = db.physical_machines_get_all(context)
        return pms

    def _db_query_by_instance_id(self, context, instance_id):
        # TODO(justinsb): Implement this properly
        pms = self._db_query_all(context)
        for pm in pms:
            if pm.instance_id == instance_id:
                return pm
        return None

    def list_instances(self):
        db_context = context.get_admin_context()
        instance_ids = []
        for pm in self._db_query_all(db_context):
            if pm.instance_id is not None:
                instance_ids.append(pm.instance_id)
        return instance_ids

    def destroy(self, instance, cleanup=True):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("destroy not supported for notsovm.")

    @exception.wrap_exception
    def attach_volume(self, instance_name, device_path, mountpoint):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("attach_volume not supported for notsovm.")

    @exception.wrap_exception
    def detach_volume(self, instance_name, mountpoint):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("detach_volume not supported for notsovm.")

    @exception.wrap_exception
    def snapshot(self, instance, image_id):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("snapshot not supported for notsovm.")

    @exception.wrap_exception
    def reboot(self, instance):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("reboot not supported for notsovm.")

    @exception.wrap_exception
    def pause(self, instance, callback):
        # This will likely never be supported
        raise exception.ApiError("pause not supported for notsovm.")

    @exception.wrap_exception
    def unpause(self, instance, callback):
        # This will likely never be supported
        raise exception.ApiError("unpause not supported for notsovm.")

    @exception.wrap_exception
    def suspend(self, instance, callback):
        # This will likely never be supported
        raise exception.ApiError("suspend not supported for notsovm")

    @exception.wrap_exception
    def resume(self, instance, callback):
        # This will likely never be supported
        raise exception.ApiError("resume not supported for notsovm")

    @exception.wrap_exception
    def rescue(self, instance, callback=None):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("rescue not supported for notsovm")

    @exception.wrap_exception
    def unrescue(self, instance, callback=None):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("rescue not supported for notsovm")

    @exception.wrap_exception
    def spawn(self, instance):
        #TODO(justinsb): Race condition in the allocation loop?
        phys_machine = None
        db_context = context.get_admin_context()
        for pm in self._db_query_all(db_context):
            if pm.instance_id is None:
                phys_machine = pm
                break

        if phys_machine is None:
            raise exception.ApiError("No available physical machines")

        net_info = PhysicalMachineNetworkInfo(phys_machine['network_config'])

        db.physical_machine_update(db_context,
                                   phys_machine['id'],
                                   {'instance_id': instance['id']})

        db.instance_set_state(db_context,
                              instance['id'],
                              power_state.NOSTATE,
                              'launching')

        ssh = self._connect_to_ssh(net_info.primary_ip)
        sftp = ssh.open_sftp()

        remote_hdd = '/dev/sda'

        # 8 Gig root, 4 gig swap, remainder available
        part_table = []
        part_table.append(',8192,L,*')
        part_table.append(',4192,S,-')
        part_table.append(',,L,-')
        part_table.append('\n')  #sfdisk is fussy about terminating lines

        LOG.debug(_("Partitioning drive: %s") % part_table)
        file = sftp.file('/tmp/parttable', mode='w')
        file.write('\n'.join(part_table))
        file.close()

        utils.ssh_execute(ssh,
                          'cat /tmp/parttable | sfdisk %s -uM' %
                          remote_hdd)

        glance_image_url = ('http://%s:%s/images/%s' %
                            (FLAGS.glance_host,
                             FLAGS.glance_port,
                             instance['image_id']))

        # Partition #1 is the image
        utils.ssh_execute(ssh,
                          'curl %s | gunzip -c > %s' %
                          (glance_image_url, remote_hdd + '1'))

        # Partition #2 is swap space
        utils.ssh_execute(ssh,
                          'mkswap %s' %
                          (remote_hdd + '2'))

        # Mount the first partition
        utils.ssh_execute(ssh,
                          'mount %s /mnt' %
                          (remote_hdd + '1'))

        # Create additional mounts for a chroot
        utils.ssh_execute(ssh,
                          'mount -t proc proc /mnt/proc')
        utils.ssh_execute(ssh,
                          'mount --bind /dev /mnt/dev')

        # Install grub
        utils.ssh_execute(ssh,
                          'grub-install --root-directory=/mnt --recheck %s' %
                          (remote_hdd))

        # Get uuids of drives
        dev_uuids = {}
        for dev in [remote_hdd + '1', remote_hdd + '2']:
            stdout, _stderr = utils.ssh_execute(ssh,
                                                'blkid -o value -s UUID %s' %
                                                (dev))
            dev_uuid = stdout.strip()
            dev_uuids[dev] = dev_uuid

        fstab = []
        fstab.append('proc\t/proc\tproc\tnodev,noexec,nosuid\t0\t0')
        fstab.append('UUID=%s\t/\text3\terrors=remount-ro\t0\t1' %
                     (dev_uuids[remote_hdd + '1']))
        fstab.append('UUID=%s\tnone\tswap\tsw\t0\t0' %
                     (dev_uuids[remote_hdd + '2']))

        LOG.debug(_("Uploading fstab: %s") % fstab)
        file = sftp.file('/mnt/etc/fstab', mode='w')
        file.write('\n'.join(fstab))
        file.close()

        # Create grub menu inside chroot
        # -y stops prompting the user if the menu.lst isn't there (!)
        utils.ssh_execute(ssh, 'chroot /mnt update-grub -y')

        net = []
        net.append('# lo0')
        net.append('auto lo')
        net.append('iface lo inet loopback')
        net.append('')
        net.append('# eth0')
        net.append('auto eth0')
        net.append('iface eth0 inet static')
        net.append('address %s' % net_info.primary_ip)
        net.append('netmask 255.255.255.0')
        net.append('gateway 10.2.1.1')

        LOG.debug(_("Uploading network config: %s") % net)
        file = sftp.file('/mnt/etc/network/interfaces', mode='w')
        file.write('\n'.join(net))
        file.close()

        utils.ssh_execute(ssh, 'mkdir -p /mnt/root/.ssh')

        key = instance['key_data']
        if key:
            LOG.debug(_("Uploading root authorized_keys: %s") % (key))
            file = sftp.file('/mnt/root/.ssh/authorized_keys', mode='w')
            file.write(key.strip() + '\n')
            file.close()
        else:
            LOG.warning(_("No SSH key data"))

        # Tell PXE to chain-boot from the hard disk
        self.pxe_manager.configure_pxe(phys_machine, net_info, 'local')

        utils.ssh_execute(ssh, 'sync')
        # We don't shut down immediately, so can we disconnect SSH cleanly
        #NOTE(justinsb): Grrr... ubuntu doesn't support anything < 1 minute
        try:
            utils.ssh_execute(ssh, 'shutdown -r now')
            ssh.close()
        except Exception as e:
            LOG.warning("Exception on shutdown (ignoring) %s" % e)

        LOG.debug(_("Physical machine initialization complete"))

        db.instance_set_state(db_context,
                              instance['id'],
                              power_state.RUNNING)

    @exception.wrap_exception
    def get_console_output(self, instance):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_console_output not supported for notsovm")

    @exception.wrap_exception
    def get_ajax_console(self, instance):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_console_output not supported for notsovm")

    def _get_instance_id(self, instance_name):
        # NOTE(justinsb): Passing around the instance name, instead of our
        #  real model PK, is total BS.  This function shouldn't be needed.
        if FLAGS.instance_name_template != 'instance-%08x':
            raise NotImplementedError("Unexpected instance_name_template")
        hexval = instance_name.replace('instance-', '')
        instance_id = int(hexval, 16)
        print "mapped %s => %s" % (instance_name, instance_id)
        return instance_id

    def get_info(self, instance_name):
        #NOTE(justinsb): Why on earth is this taking the instance_name?
        instance_id = self._get_instance_id(instance_name)

        db_context = context.get_admin_context()
        pm = self._db_query_by_instance_id(db_context, instance_id)
        if pm is None:
            raise exception.NotFound(_("Instance %s not found")
                                     % instance_name)

        #(state, max_mem, mem, num_cpu, cpu_time) = virt_dom.info()
        return {'state': power_state.RUNNING,
                'max_mem': 0,
                'mem': 0,
                'num_cpu': 0,
                'cpu_time': 0}

    def get_diagnostics(self, instance_name):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("diagnostics not supported for notsovm")

    def get_disks(self, instance_name):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_disks not supported for notsovm")

    def get_interfaces(self, instance_name):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_interfaces not supported for notsovm")

    def get_vcpu_total(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_vcpu_total not supported for notsovm")

    def get_memory_mb_total(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_memory_mb_total not supported for notsovm")

    def get_local_gb_total(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_local_gb_total not supported for notsovm")

    def get_vcpu_used(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_vcpu_used not supported for notsovm")

    def get_memory_mb_used(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_memory_mb_used not supported for notsovm")

    def get_local_gb_used(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_local_gb_used not supported for notsovm")

    def get_hypervisor_type(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_hypervisor_type not supported for notsovm")

    def get_hypervisor_version(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_hypervisor_version not supported for notsovm")

    def get_cpu_info(self):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_cpu_info not supported for notsovm")

    def block_stats(self, instance_name, disk):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("block_stats not supported for notsovm")

    def interface_stats(self, instance_name, interface):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("interface_stats not supported for notsovm")

    def get_console_pool_info(self, console_type):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("get_console_pool_info not supported for notsovm")

    def refresh_security_group_rules(self, security_group_id):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("refresh_security_group_rules not supported for notsovm")

    def refresh_security_group_members(self, security_group_id):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("refresh_security_group_members not supported for notsovm")

    def compare_cpu(self, cpu_info):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("compare_cpu not supported for notsovm")

    def ensure_filtering_rules_for_instance(self, instance_ref):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("ensure_filtering_rules_for_instance not supported for notsovm")

    def live_migration(self, ctxt, instance_ref, dest,
                       post_method, recover_method):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("live_migration not supported for notsovm")

    def unfilter_instance(self, instance_ref):
        LOG.warning("Hit unhandled method")
        raise exception.ApiError("unfilter_instance not supported for notsovm")


    def _connect_to_ssh(self, ip):
        ssh = paramiko.SSHClient()
        #TODO(justinsb): We need a better SSH key policy
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if FLAGS.notso_privatekey:
            privatekeyfile = os.path.expanduser(FLAGS.notso_privatekey)
            # It sucks that paramiko doesn't support DSA keys
            privatekey = paramiko.RSAKey.from_private_key_file(privatekeyfile)
            ssh.connect(ip,
                        port=22,
                        username='root',
                        pkey=privatekey)
        else:
            raise exception.Error(_("Specify notso_privatekey"))
        return ssh


